/*
   HappyMod 2
   v 25034
*/

// include
//--------------------------------------------
#define DIRECTINPUT_VERSION   0x0800
#include <windows.h>
#include <dinput.h>
#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include <float.h>
#include "Log.h"
#include "share.h"
#include "hook.h"
#include "m.h"


// constants
//--------------------------------------------
#define CRC_FILE                    "crc.txt"
#define CRC_LIMIT                   1000
#define HM2_VL2_FILE                "GameData\\base\\_hm2.vl2"

#define HISTORY_SIZE                13
#define BEACON_MAX                  32
#define OBJECT_MAX                  64
#define PLAYER_MAX                  64
#define TARGET_MAX                  (OBJECT_MAX+PLAYER_MAX)

#define TEAM_OBSERVER               0
#define TEAM_INFERNO                1
#define TEAM_STORM                  2
#define TEAM_OTHER                  69

#define WEAPON_OTHER                0
#define WEAPON_SNIPER               1
#define WEAPON_SHOCKLANCE           2
#define WEAPON_LAST                 2

#define TRIGGER                     "friendBeaconLineEndColor"
#define TRIGGER_AA_FIRE             100
#define TRIGGER_AA_TWEAK            101
#define TRIGGER_AA_TEAMMATES        102
#define TRIGGER_AA_FIREDELAY        103
#define TRIGGER_AA_WEAPON           104
#define TRIGGER_AA_SLDISTANCE       105

#define BEACON_REFRESH              20
#define BEACON_OTHER                0
#define BEACON_BEACON               1
#define BEACON_DEPLOYABLE           2

#define AA_TWEAK_MIN                0.5f
#define AA_TWEAK_MAX                30.0f
#define AA_FIREDELAY_MIN            1.0f
#define AA_FIREDELAY_MAX            1000.0f
#define AA_SLDISTANCE_MIN           1.0f
#define AA_SLDISTANCE_MAX           1000.0f
#define AA_SENSITIVITY              0.8f


// local types
//--------------------------------------------
struct sPatch
{
   const char*    description;
   DWORD          address;
   unsigned char* patch;
   unsigned char* backup;
   DWORD          size;
};

struct sHook
{
   const char*    description;
   DWORD          address;
   fHookHandler   handler;
};

struct sCrcMap
{
   DWORD          crc_real[2];
   DWORD          crc_fake[2];
};

struct sBeaconEntry
{
   bool           active;
   int            id;
   sVertex3f      original_pos;
   float*         p_x;
   float*         p_y;
   float*         p_z;
   int*           p_type;
};

struct sObjectEntry
{
   bool           active;
   int            id;
   float*         p_x;
   float*         p_y;
   float*         p_z;
   int            team_id;
   int            beacon_type;
};

struct sPlayer
{
   bool           active;
   int            id;
   float*         p_x;
   float*         p_y;
   float*         p_z;
   int            team_id;
   sVertex3f      position_history[HISTORY_SIZE];
};

struct sPlayerSelf
{
   int*           id;
   float*         p_x;
   float*         p_y;
   float*         p_z;
   float*         ori_x;
   float*         ori_y;
   float*         pitch;
};


// tribes2 types
//--------------------------------------------
struct sEntityObjectInfo
{
   char*                description;   // 0x0024
};

struct sEntityShapeInfo2
{
   PAD(0x0044)
   char*                shape_file;    // 0x0044
};

struct sEntityShapeInfo1
{
   PAD(0x01fc)
   sEntityShapeInfo2*   shape_info_2;  // 0x01fc
};

struct sEntityPosition
{
   float                x;             // 0x0000
   PAD(0x000c)
   float                y;             // 0x0010
   PAD(0x000c)
   float                z;             // 0x0020
};

struct sEntityPositionWP
{
   float                x;             // 0x0000
   float                y;             // 0x0004
   float                z;             // 0x0008
};

struct sEntityTeamInfo
{
   PAD(0x0018)
   int                  team_id;       // 0x0018
};

struct sEntity
{
   PAD(0x0020)
   int                  object_id;     // 0x0020
   sEntityObjectInfo*   object_info;   // 0x0024
   PAD(0x0010)
   sEntityPositionWP    position_wp;   // 0x0038
   PAD(0x0008)
   sEntityShapeInfo1*   shape_info_1;  // 0x004c
   PAD(0x004c)
   float                ori_y;         // 0x009c
   float                ori_x;         // 0x00a0
   PAD(0x0004)
   sEntityPosition      position;      // 0x00a8
   PAD(0x01a0)
   sEntityTeamInfo*     team_info;     // 0x026c
   PAD(0x0618)
   int                  object_id_2;   // 0x0888
   PAD(0x00b4)
   float                pitch;         // 0x0940
};

struct sBeaconTarget
{
   float*   p_x;
   float*   p_y;
   float*   p_z;
   float    yaw;
   int      beacon_type;
   bool*    active;
};


// local prototypes
//--------------------------------------------
static bool appEnableCheats( void );
static bool appReset( void );

static HOOK_CALL hookCrc( sRegisters reg );
static HOOK_CALL hookCreateObject( sRegisters reg );
static HOOK_CALL hookDestroyObject( sRegisters reg );
static HOOK_CALL hookCreateWaypoint( sRegisters reg );
static HOOK_CALL hookDestroyWaypoint( sRegisters reg );
static HOOK_CALL hookChangeTeams( sRegisters reg );
static HOOK_CALL hookDinput( sRegisters reg );
static HOOK_CALL hookFrame( sRegisters reg );
static HOOK_CALL hookPing( sRegisters reg );
static HOOK_CALL hookFps( sRegisters reg );
static HOOK_CALL hookTrigger( sRegisters reg );
static HOOK_CALL hookBeaconFix( sRegisters reg );

static bool togglePlayerFinder( sRegisters &reg );
static bool toggleWhiteOut( sRegisters &reg );
static bool toggleCloak( sRegisters &reg );
static bool toggleAutoBeacon( sRegisters &reg );

static bool beaconInit( void );
static bool beaconRefresh( void );

static void aaExecute( void );
static void aaGetOrientationToTarget( float& out_yaw, float& out_pitch, sVertex3f& in_target );

static void scriptCall( const char* command, ... );
static void scriptMessage( const char* message, ... );
static void scriptEcho( const char* message, ... );
static bool loadCrcs( void );
static bool checkFile( const char* file );


// patches
//--------------------------------------------

// patches
static unsigned char       g_patch_fakecrc[] =           { 0xeb };
static unsigned char       g_patch_vistdist[] =          { 0xeb };
static unsigned char       g_patch_vistdistcheck[] =     { 0xe4 };
static unsigned char       g_patch_whiteout[] =          { 0xb8, 0x00, 0x00, 0x00, 0x00, 0x90 };
static unsigned char       g_patch_cloak[] =             { 0xb8, 0x66, 0x66, 0x66, 0x3f };
static unsigned char       g_patch_ffialways[] =         { 0x90, 0x90 };
static unsigned char       g_patch_ffifar[] =            { 0xb8, 0x00, 0x40, 0x9c, 0x46, 0x90 };
static unsigned char       g_patch_globalbeacons[] =     { 0x90, 0x90 };

// patch backups
static unsigned char       g_backup_fakecrc[] =          { 0x74 };
static unsigned char       g_backup_vistdist[] =         { 0x75 };
static unsigned char       g_backup_vistdistcheck[] =    { 0xd0 };
static unsigned char       g_backup_whiteout[] =         { 0x8b, 0x85, 0xac, 0xfe, 0xff, 0xff };
static unsigned char       g_backup_cloak[] =            { 0xa1, 0x84, 0x73, 0x9a, 0x00 };
static unsigned char       g_backup_ffialways[] =        { 0x74, 0x09 };
static unsigned char       g_backup_ffifar[] =           { 0x8b, 0x82, 0x2c, 0x0e, 0x00, 0x00 };
static unsigned char       g_backup_globalbeacons[] =    { 0x75, 0x39 };

// patch indexes
#define  PATCH_INDEX_VISDIST           0
#define  PATCH_INDEX_VISDISTCHECK      1
#define  PATCH_INDEX_WHITEOUT          2
#define  PATCH_INDEX_CLOAK             3
#define  PATCH_INDEX_FFIALWAYS         4
#define  PATCH_INDEX_FFIFAR            5
#define  PATCH_INDEX_FAKECRC           6
#define  PATCH_INDEX_GLOBALBEACONS     7

// tribes2.exe function addersses
#define  CALL_SCRIPT                   0x00426690


// globals
//--------------------------------------------
// patches
static sPatch        g_patches[] = {
   { "vis dist",           0x00569c2f,    g_patch_vistdist,       g_backup_vistdist,         1},
   { "vis dist check",     0x00569e60,    g_patch_vistdistcheck,  g_backup_vistdistcheck,    1},
   { "no whiteout",        0x005fbc17,    g_patch_whiteout,       g_backup_whiteout,         6},
   { "no cloak",           0x006a56f4,    g_patch_cloak,          g_backup_cloak,            5},
   { "ffi always",         0x00503ff5,    g_patch_ffialways,      g_backup_ffialways,        2},
   { "ffi far",            0x005bbd2a,    g_patch_ffifar,         g_backup_ffifar,           6},
   { "global beacons",     0x00689e2e,    g_patch_globalbeacons,  g_backup_globalbeacons,    2},
   { "fake crc",           0x005e4618,    g_patch_fakecrc,        g_backup_fakecrc,          1},
};

// hooks
static sHook         g_hooks[] = {
   { "crc",                0x004408a8,    hookCrc },
   { "create object",      0x0058c200,    hookCreateObject },
   { "destroy object",     0x004390a8,    hookDestroyObject },
   { "create waypoint",    0x00671834,    hookCreateWaypoint },
   { "destroy waypoint",   0x00671620,    hookDestroyWaypoint },
   { "change teams",       0x00584461,    hookChangeTeams },
   { "dinput",             0x00558978,    hookDinput },
   { "frame",              0x0056457f,    hookFrame },
   { "ping",               0x0050cbf7,    hookPing },
   { "fps",                0x005c4644,    hookFps },
   { "trigger",            0x004295db,    hookTrigger },
   { "beacon fix",         0x005f4610,    hookBeaconFix },
};

// crc map
static int           g_crc_count = 0;
static sCrcMap       g_crc_map[CRC_LIMIT] = {0};



// other globals
//--------------------------------------------
float             g_fps_history[HISTORY_SIZE];
float             g_ping_history[HISTORY_SIZE];
sBeaconEntry      g_beacons[BEACON_MAX];
sObjectEntry      g_objects[OBJECT_MAX];
sPlayer           g_players[PLAYER_MAX];
sPlayerSelf       g_me;
int               g_team_id;
float             g_ping;
float             g_fps;
bool              g_autobeacon_enable;
float             g_autoaim_tweak;
float             g_autoaim_firedelay;
float             g_autoaim_sldistance;
bool              g_autoaim_teammates;
int               g_weapon;
static char       g_path[BUF_PATH_SIZE]= "roar";


// app impl
//--------------------------------------------
bool appStart( sInitPDU& pdu )
{
   appReset();
   if(!loadCrcs())
   {
      print("warning: %s not found", CRC_FILE);
      g_crc_count = 0;
   }

   if(!appEnableCheats())
      return false;
   print("patched");

   return true;
}


//--------------------------------------------
void appStop( void )
{
   int   i;

   // unload hooks
   for(i=0; i<ARRAY_SIZE(g_hooks); i++)
      if( !hookRemove(g_hooks[i].address, g_hooks[i].handler) )
         return;

   // unload patches
   for(i=0; i<ARRAY_SIZE(g_patches); i++)
      memcpy((void*)g_patches[i].address, g_patches[i].backup, g_patches[i].size);
   print("restored");

   // reset to original state
   appReset();
}


//--------------------------------------------
static bool appEnableCheats( void )
{
   int  i;

   // load hooks
   for(i=0; i<ARRAY_SIZE(g_hooks); i++)
      if( !hookSet(g_hooks[i].address, g_hooks[i].handler) )
         return false;
   // load patches
   for(i=0; i<ARRAY_SIZE(g_patches); i++)
   {
      DWORD    old_permissions;
      if( !VirtualProtect((void *)g_patches[i].address, g_patches[i].size, PAGE_EXECUTE_READWRITE, &old_permissions) )
         return false;
      memcpy((void*)g_patches[i].address, g_patches[i].patch, g_patches[i].size);
   }
   // reset beacons
   if(!beaconInit())
      return false;

   return true;
}


//--------------------------------------------
static bool appReset( void )
{
   strcpy(g_path, g_init_pdu.path);
   memset(g_fps_history, 0, ARRAY_SIZE(g_fps_history)*sizeof(g_fps_history[0]));
   memset(g_ping_history, 0, ARRAY_SIZE(g_ping_history)*sizeof(g_ping_history[0]));
   memset(g_beacons, 0, ARRAY_SIZE(g_beacons)*sizeof(g_beacons[0]));
   memset(g_objects, 0, ARRAY_SIZE(g_objects)*sizeof(g_objects[0]));
   memset(g_players, 0, ARRAY_SIZE(g_players)*sizeof(g_players[0]));
   memset(&g_me, 0, sizeof sPlayer);
   g_team_id = TEAM_OBSERVER;
   g_ping = 0.0f;
   g_fps = 60.0f;
   g_autobeacon_enable = true;
   g_autoaim_tweak = 7.0f;
   g_autoaim_firedelay = 50.0f;
   g_autoaim_sldistance = 40.0f;
   g_autoaim_teammates = true;
   g_weapon = WEAPON_OTHER;

   return true;
}


// hooks
//-------------------------------------------
static HOOK_CALL hookCrc( sRegisters reg )
{
   DWORD    crc[2];
   int      i;

   // good time to reset beacons
   if(!beaconInit())
      return;

   // at this point a final 8 byte crc is calculated for all models in question which varies
   // per server. this goes through the crc map to find the valid crc given the invalid one
   // calculated from _hm2.vl2
   memcpy(crc, (void*)reg.ebx, sizeof(crc));
   for(i=0; i<g_crc_count; i++)
   {
      if(crc[0]==g_crc_map[i].crc_fake[0] && crc[1]==g_crc_map[i].crc_fake[1])
      {
         memcpy((void*)reg.ebx, g_crc_map[i].crc_real, sizeof(crc));
         print("event: crc spoof");
         return;
      }
   }

   print("failed: crc spoof");
   g_log.print("%s: 0x%0.8x,0x%0.8x\n", (checkFile(HM2_VL2_FILE))?"CRC FAKE":"CRC REAL", crc[0], crc[1]);
}


//--------------------------------------------
static HOOK_CALL hookCreateObject( sRegisters reg )
{
   char*       description;
   sEntity*    entity;

   entity = (sEntity *)reg.ebx;
   if(entity==NULL || entity->object_info==NULL || entity->object_info->description==NULL)
      return;
   description = entity->object_info->description;

   if(strcmp(description,"Player")==0)
   {
      // find free player
      for(int i=0; i<PLAYER_MAX; i++)
         if(g_players[i].active == false)
         {
            g_players[i].id = entity->object_id;
            g_players[i].p_x = &entity->position.x;
            g_players[i].p_y = &entity->position.y;
            g_players[i].p_z = &entity->position.z;

            // read team
            if(entity->team_info == NULL)
               g_players[i].team_id = TEAM_OTHER;
            else
               g_players[i].team_id = entity->team_info->team_id;
            g_players[i].active = true;

            break;
         }
   }
   else if(strcmp(description,"BeaconObject")==0)
   {
      sBeaconEntry*     b;

      // find free beacon
      for(int i=0; i<BEACON_MAX; i++)
         if(g_beacons[i].active == false)
            break;
      if(i == BEACON_MAX)
         return;
      b = &g_beacons[i];

      // read object info
      b->id = entity->object_id;
      b->p_x = &entity->position.x;
      b->p_y = &entity->position.y;
      b->p_z = &entity->position.z;
      b->original_pos.x = entity->position.x;
      b->original_pos.y = entity->position.y;
      b->original_pos.z = entity->position.z;
      b->active = true;

      // set new type
      entity->object_id_2 = BEACON_BEACON;
      b->p_type = &entity->object_id_2;
      if(!g_autobeacon_enable)
         *b->p_type = BEACON_OTHER;
   }
   else if(strcmp(description,"Turret")==0 || strcmp(description,"StaticShape")==0)
   {
      sObjectEntry*     o;
      char*             shape_file;

      // find free object
      for(int i=0; i<OBJECT_MAX; i++)
         if(g_objects[i].active == false)
            break;
      if(i == OBJECT_MAX)
         return;
      o = &g_objects[i];

      // check for model name
      if(entity->shape_info_1==NULL || entity->shape_info_1->shape_info_2==NULL || entity->shape_info_1->shape_info_2->shape_file==NULL)
         return;
      shape_file = entity->shape_info_1->shape_info_2->shape_file;

      // only those listed here get beaconed
      if(strcmp(shape_file,"deploy_inventory.dts")==0 ||
         strcmp(shape_file,"turret_outdoor_deploy.dts")==0 ||
         strcmp(shape_file,"turret_indoor_deployc.dts")==0 ||
         strcmp(shape_file,"turret_indoor_deployf.dts")==0 ||
         strcmp(shape_file,"turret_indoor_deployw.dts")==0 )
      {
         o->beacon_type = BEACON_DEPLOYABLE;
      }
      else if(
         strcmp(shape_file,"turret_base_large.dts")==0 ||
         strcmp(shape_file,"sensor_pulse_large.dts")==0 ||
         strcmp(shape_file,"ext_flagstand.dts")==0 ||
         strcmp(shape_file,"vehicle_pad_station.dts")==0)
      {
         o->beacon_type = BEACON_OTHER;
      }
      else
         return;

      // read team
      if(entity->team_info == NULL)
         o->team_id = TEAM_OTHER;
      else
         o->team_id = entity->team_info->team_id;

      // read object info
      o->id = entity->object_id;
      o->p_x = &entity->position.x;
      o->p_y = &entity->position.y;
      o->p_z = &entity->position.z;
      o->active = true;
   }
}

//--------------------------------------------
static HOOK_CALL hookDestroyObject( sRegisters reg )
{
   int   i;
   int   id = reg.ebx;

   for(i=0; i<PLAYER_MAX; i++)
   {
      // destroy player
      if(g_players[i].id == id)
      {
         g_players[i].id = 0;
         g_players[i].active = false;
         return;
      }
   }

   for(i=0; i<BEACON_MAX; i++)
   {
      // destroy beacon
      if(g_beacons[i].id == id)
      {
         g_beacons[i].id = 0;
         g_beacons[i].active = false;
         return;
      }
   }
   
   for(i=0; i<OBJECT_MAX; i++)
   {
      // destroy object
      if(g_objects[i].id == id)
      {
         g_objects[i].id = 0;
         g_objects[i].active = false;
         return;
      }
   }
}


//--------------------------------------------
static HOOK_CALL hookCreateWaypoint( sRegisters reg )
{
   sEntity*          entity;
   sObjectEntry*     o;

   entity = (sEntity *)reg.ebx;

   // find free object
   for(int i=0; i<OBJECT_MAX; i++)
      if(g_objects[i].active == false)
         break;
   if(i == OBJECT_MAX)
      return;
   o = &g_objects[i];

   // create obj
   o->id = entity->object_id;
   o->p_x = &entity->position_wp.x;
   o->p_y = &entity->position_wp.y;
   o->p_z = &entity->position_wp.z;
   o->active = true;
   o->team_id = TEAM_OTHER;
   o->beacon_type = BEACON_OTHER;
}


//--------------------------------------------
static HOOK_CALL hookDestroyWaypoint( sRegisters reg )
{
   sEntity*    entity;

   entity = (sEntity *)reg.ebx;

   // find waypoint
   for(int i=0; i<OBJECT_MAX; i++)
   {
      if(g_objects[i].id == entity->object_id)
      {
         g_objects[i].id = 0;
         g_objects[i].active = false;
         return;
      }
   }
}


//--------------------------------------------
static HOOK_CALL hookChangeTeams( sRegisters reg )
{
   sEntity*    entity = (sEntity*)reg.esi;

   if(entity)
   {
      // get team id
      if(entity->team_info)
      {
         g_team_id = entity->team_info->team_id;
         g_me.id = &entity->object_id;
         g_me.p_x = &entity->position.x;
         g_me.p_y = &entity->position.y;
         g_me.p_z = &entity->position.z;
         g_me.ori_x = &entity->ori_x;
         g_me.ori_y = &entity->ori_y;
         g_me.pitch = &entity->pitch;
      }
      else
      {
         g_team_id = TEAM_OBSERVER;
      }
   }
}

//--------------------------------------------
static HOOK_CALL hookDinput( sRegisters reg )
{
   switch(reg.eax+1)
   {
   case DIK_F10:
      togglePlayerFinder(reg);
      break;
   case DIK_F11-2:
      toggleWhiteOut(reg);
      break;
   case DIK_F12-2:
      toggleAutoBeacon(reg);
      break;
   }
}


//--------------------------------------------
static HOOK_CALL hookFrame( sRegisters reg )
{
   if(g_team_id)
   {
      // save player history
      for(int i=0; i<PLAYER_MAX; i++)
      {
         if(!g_players[i].active)
            continue;
         for(int j=0; j<(HISTORY_SIZE-1); j++)
            g_players[i].position_history[j] = g_players[i].position_history[j+1];
         if(g_players[i].p_x && g_players[i].p_y && g_players[i].p_z)
         {
            g_players[i].position_history[j].x = *g_players[i].p_x;
            g_players[i].position_history[j].y = *g_players[i].p_y;
            g_players[i].position_history[j].z = *g_players[i].p_z;
         }
      }

      // save fps and ping history
      for(i=0; i<(HISTORY_SIZE-1); i++)
      {
         g_fps_history[i] = g_fps_history[i+1];
         g_ping_history[i] = g_ping_history[i+1];
      }
      g_fps_history[i] = g_fps;
      g_ping_history[i] = g_ping;

      // do autobeacon
      if(g_autobeacon_enable)
         beaconRefresh();
   }
}


//--------------------------------------------
static HOOK_CALL hookPing( sRegisters reg )
{
   if(reg.eax)
   {
      const char* pingstr = (const char*)reg.eax;
      sscanf(pingstr, "%f", &g_ping);
      if(ISNAN(g_ping) || g_ping<=0.5f || g_ping>5000.0f)
         g_ping = 0.5f;
   }
}


//-------------------------------------------
static HOOK_CALL hookFps( sRegisters reg )
{
   if(reg.eax)
   {
      const char* fpsstr = (const char*)reg.eax;
      sscanf(fpsstr, "%f", &g_fps);
      if(ISNAN(g_fps) || g_fps<=0.1f || g_fps>5000.0f)
         g_fps = 60.0f;
   }
}


//-------------------------------------------
static HOOK_CALL hookTrigger( sRegisters reg )
{
   // external script trigger
   if(reg.edi && *(int *)reg.edi && strcmp((char *)(*(DWORD *)reg.edi),TRIGGER)==0)
   {
      int    f1 = (int)(*(float*)(reg.esi));
      float& f2 = *(float*)(reg.esi+4);
      float& f3 = *(float*)(reg.esi+8);

      switch(f1)
      {
      case TRIGGER_AA_FIRE:
         // do autoaim
         aaExecute();
         break;
      case TRIGGER_AA_TWEAK:
         // autoaim tweak
         if(f2<AA_TWEAK_MIN)
            g_autoaim_firedelay = AA_TWEAK_MIN;
         else if(f2>AA_TWEAK_MAX)
            g_autoaim_firedelay = AA_TWEAK_MAX;
         else
            g_autoaim_tweak = f2;
         break;
      case TRIGGER_AA_FIREDELAY:
         // fire delay
         if(f2<AA_FIREDELAY_MIN)
            g_autoaim_firedelay = AA_FIREDELAY_MIN;
         else if(f2>AA_FIREDELAY_MAX)
            g_autoaim_firedelay = AA_FIREDELAY_MAX;
         else
            g_autoaim_firedelay = f2;
         break;
      case TRIGGER_AA_TEAMMATES:
         // autoaim teammates targeting toggle
         if(f2)
            g_autoaim_teammates = true;
         else
            g_autoaim_teammates = false;
         break;
      case TRIGGER_AA_WEAPON:
         // select weapon type
         g_weapon = (int)f2;
         if(g_weapon > WEAPON_LAST)
            g_weapon = WEAPON_OTHER;
         break;
      case TRIGGER_AA_SLDISTANCE:
         // shocklance distance
         if(f2<AA_SLDISTANCE_MIN)
            g_autoaim_sldistance = AA_SLDISTANCE_MIN;
         else if(f2>AA_SLDISTANCE_MAX)
            g_autoaim_sldistance = AA_SLDISTANCE_MAX;
         else
            g_autoaim_sldistance = f2;
         
         break;
      }
   }
}

//-------------------------------------------
static HOOK_CALL hookBeaconFix( volatile sRegisters reg )
{
    switch(reg.eax)
    {
    case 0x0000444:
        reg.eax = 0x000015c;
        break;
    case 0x0000441:
        reg.eax = 0x0000186;
        break;
    case 0x000043b:
        reg.eax = 0x00001ce;
        break;
    }
}

// toggles
//-------------------------------------------
static bool togglePlayerFinder( sRegisters &reg )
{
   static bool    up = 0;
   static bool    toggle = 0;
   sPatch*        patch1 = &g_patches[PATCH_INDEX_FFIALWAYS];
   sPatch*        patch2 = &g_patches[PATCH_INDEX_FFIFAR];

   if(up^=1)
   {
      if(toggle^=1)
      {
         memcpy((void*)patch1->address, patch1->backup, patch1->size);
         memcpy((void*)patch2->address, patch2->backup, patch2->size);
      }
      else
      {
         memcpy((void*)patch1->address, patch1->patch, patch1->size);
         memcpy((void*)patch2->address, patch2->patch, patch2->size);
      }
      scriptMessage("%s: %s", patch1->description, toggle?"OFF":"ON");
   }
   return true;
}

//-------------------------------------------
static bool toggleWhiteOut( sRegisters &reg )
{
   static bool    up = 0;
   static bool    toggle = 0;
   sPatch*        patch1 = &g_patches[PATCH_INDEX_WHITEOUT];

   if(up^=1)
   {
      if(toggle^=1)
         memcpy((void*)patch1->address, patch1->backup, patch1->size);
      else
         memcpy((void*)patch1->address, patch1->patch, patch1->size);
      scriptMessage("%s: %s", patch1->description, toggle?"OFF":"ON");
   }
   return true;
}

//-------------------------------------------
static bool toggleCloak( sRegisters &reg )
{
   static bool    up = 0;
   static bool    toggle = 0;
   sPatch*        patch1 = &g_patches[PATCH_INDEX_CLOAK];

   if(up^=1)
   {
      if(toggle^=1)
         memcpy((void*)patch1->address, patch1->backup, patch1->size);
      else
         memcpy((void*)patch1->address, patch1->patch, patch1->size);
      scriptMessage("%s: %s", patch1->description, toggle?"OFF":"ON");
   }
   return true;
}

//-------------------------------------------
static bool toggleAutoBeacon( sRegisters &reg )
{
   static bool    up = 0;
   static bool    toggle = 0;

   if(up^=1)
   {
      if(g_autobeacon_enable)
      {
         g_autobeacon_enable = false;
         for(int i=0; i<BEACON_MAX; i++)
         {
            // restore original beacon positions
            if(g_beacons[i].active==true)
            {
               *g_beacons[i].p_x = g_beacons[i].original_pos.x;
               *g_beacons[i].p_y = g_beacons[i].original_pos.y;
               *g_beacons[i].p_z = g_beacons[i].original_pos.z;
               *g_beacons[i].p_type = BEACON_OTHER;
            }
         }
      }
      else
         g_autobeacon_enable = true;
      scriptMessage("autobeacon: %s", g_autobeacon_enable?"ON":"OFF");
   }
   return true;
}


// autobeacon
//-------------------------------------------
static bool beaconInit( void )
{
   memset(g_beacons, 0, sizeof(g_beacons));
   memset(g_objects, 0, sizeof(g_objects));
   return true;
}

//-------------------------------------------
int compareBeaconTarget( const void *elem1, const void *elem2 )
{
   return (((sBeaconTarget*)elem1)->yaw > ((sBeaconTarget*)elem2)->yaw)?1:-1;
}

//-------------------------------------------
static bool beaconRefresh( void )
{
   static unsigned int     frame = 0;
   static sBeaconTarget    beacon_targets[TARGET_MAX];
   static int              target_count = 0;
   int                     i, j;

   // dont need to do this every frame
   if((frame++%BEACON_REFRESH) == 0)
   {
      // get targets
      target_count = 0;
      for(i=0; i<OBJECT_MAX; i++)
         if(g_objects[i].active && g_objects[i].team_id>0 && (g_objects[i].team_id!=g_team_id || g_objects[i].team_id==TEAM_OTHER))
         {
            beacon_targets[target_count].p_x = g_objects[i].p_x;
            beacon_targets[target_count].p_y = g_objects[i].p_y;
            beacon_targets[target_count].p_z = g_objects[i].p_z;
            beacon_targets[target_count].beacon_type = g_objects[i].beacon_type;
            beacon_targets[target_count].active = &g_objects[i].active;
            target_count++;
         }
      // get players
      for(i=0; i<PLAYER_MAX; i++)
         if(g_players[i].active && g_players[i].team_id!=g_team_id && *g_me.id!=g_players[i].id)
         {
            beacon_targets[target_count].p_x = g_players[i].p_x;
            beacon_targets[target_count].p_y = g_players[i].p_y;
            beacon_targets[target_count].p_z = g_players[i].p_z;
            beacon_targets[target_count].beacon_type = BEACON_OTHER;
            beacon_targets[target_count].active = &g_players[i].active;
            target_count++;
         }
      // calculate distance to center of screen
      for(i=0; i<target_count; i++)
      {
         float       pitch;
         sVertex3f   target;
         target.x =  *beacon_targets[i].p_x;
         target.y =  *beacon_targets[i].p_y;
         target.z =  *beacon_targets[i].p_z;
         aaGetOrientationToTarget(beacon_targets[i].yaw, pitch, target);
         beacon_targets[i].yaw = (float)fabs(beacon_targets[i].yaw);
      }

      // sort by distance
      qsort(beacon_targets, target_count, sizeof sBeaconTarget, compareBeaconTarget);
   }

   // assign to beacons
   for(i=0, j=0; i<BEACON_MAX && j<target_count; i++)
   {
      if(g_beacons[i].active==true && *beacon_targets[i].active)
      {
         *g_beacons[i].p_x = *beacon_targets[j].p_x;
         *g_beacons[i].p_y = *beacon_targets[j].p_y;
         *g_beacons[i].p_z = *beacon_targets[j].p_z;
         *g_beacons[i].p_type = beacon_targets[j].beacon_type;
         j++;
      }
   }
   // free any unassigned beacons
   for(; i<BEACON_MAX; i++)
   {
      if(g_beacons[i].active==true)
      {
         *g_beacons[i].p_x = g_beacons[i].original_pos.x;
         *g_beacons[i].p_y = g_beacons[i].original_pos.y;
         *g_beacons[i].p_z = g_beacons[i].original_pos.z;
         *g_beacons[i].p_type = BEACON_BEACON;
      }
   }

   return true;
}


// auto aim
//-------------------------------------------
static float aaGetLead( void )
{
   int   index = HISTORY_SIZE-1;
   float fps = g_fps_history[index];	   // last fps
   float ping = g_ping_history[index] / 1000.0f;   // last ping
   float delay = g_autoaim_firedelay / 1000.0f;    // fire delay
   float lead = (ping + delay + g_autoaim_tweak/10.0f) / ((float)HISTORY_SIZE / fps);
   scriptMessage("aa lead: %0.2f (ping=%0.2f delay=%0.2f fps=%0.2f)", lead, ping, delay, fps);

   return lead;
}

//-------------------------------------------
static void aaGetNonLinearTarget( sVertex3f& vt, sPlayer& player )
{
   // a simplified reduction from least squares
   // to accomidate typical t2 movement

   sVertex3f   v[3];
   float       lead;
   
   v[0] = player.position_history[0];
   v[1] = player.position_history[HISTORY_SIZE/2];
   v[2] = player.position_history[HISTORY_SIZE-1];
   lead = aaGetLead();

   // lead vars
   sVertex3f   v1, v2;
   sVertex3f   a1, a2, af;
   float       m1, m2, ma, m;

   // get 2 vectors of curve
   vecSubtract3( v1, v[1], v[0] );
   vecSubtract3( v2, v[2], v[1] );

   // magnitude and angles of vectors
   m1 = vecDirCos3(a1,v1);
   m2 = vecDirCos3(a2,v2);

   // get angle between vectors
   af.x = a2.x - a1.x;
   af.y = a2.y - a1.y;
   af.z = a2.z - a1.z;
   
   // restrict bend to 90 degrees
   RANGE_TO(af.x, PID2);
   RANGE_TO(af.y, PID2);
   RANGE_TO(af.z, PID2);
      
   // pred angle factor
   af.x = a1.x + af.x + af.x;
   af.y = a1.y + af.y + af.y;
   af.z = a1.z + af.z + af.z;

   // determine final magnitude
   if(m2 < m1)
      m = m2 * lead * (float)pow(m2/m1, lead);
   else
      m = m2 * lead;

   // get final coordinate
   af.x = (float)cos(af.x);
   af.y = (float)cos(af.y);
   af.z = (float)cos(af.z);
   ma = vecMagnitude3(af);
   if(ISNAN(m))
      m = 0.0f;
   vt.x = v[2].x + m * af.x / ma;
   vt.y = v[2].y + m * af.y / ma;
   vt.z = v[2].z + m * af.z / ma;
}


//-------------------------------------------
static void aaGetInstantTarget( sVertex3f& vt, sPlayer& player )
{
   vt = player.position_history[HISTORY_SIZE-1];
}

//-------------------------------------------
static void aaGetOrientationToTarget( float& out_yaw, float& out_pitch, sVertex3f& in_target )
{
   float       f;
   sVertex3f   pos_me;
   sVertex3f   vectorto_target;
   sVertex3f   orientation_me;

   // init vars
   in_target = in_target;

   pos_me.x = *g_me.p_x;
   pos_me.y = *g_me.p_y;
   pos_me.z = *g_me.p_z;

   // get vector to target
   vecSubtract3(vectorto_target, in_target, pos_me);

   // make unit vector
   f = vecMagnitude2((sVertex2f&)vectorto_target);
   vectorto_target.z /= vecMagnitude3(vectorto_target);
   vectorto_target.z = (float)asin(vectorto_target.z);
   vectorto_target.x /= f;
   vectorto_target.y /= f;

   // get self vector
   orientation_me.x = *g_me.ori_x;
   orientation_me.y = *g_me.ori_y;
   orientation_me.z = -*g_me.pitch;

   // polarize self vector
   out_yaw = vecAngle2((sVertex2f&)vectorto_target, (sVertex2f&)orientation_me);
   out_pitch = orientation_me.z - vectorto_target.z;

   // if nan's then we're right on target
   if(ISNAN(out_yaw))
      out_yaw = 0.0f;
   if(ISNAN(out_pitch))
      out_pitch = 0.0f;
}

//-------------------------------------------
static float aaGetDistanceToTarget( sVertex3f& in_target )
{
   float       m;
   sVertex3f   pos_me;
   sVertex3f   vectorto_target;

   pos_me.x = *g_me.p_x;
   pos_me.y = *g_me.p_y;
   pos_me.z = *g_me.p_z;

   vecSubtract3(vectorto_target, in_target, pos_me);
   m = vecMagnitude3(vectorto_target);

   return m;
}


//-------------------------------------------
static void aaExecute( void )
{
   int         i = 0;
   int         final_index = -1;
   float       final_yaw = 100.0f;
   float       final_pitch = 100.0f;
   float       mag = 100.0f;

   // ignore if not on a team
   if(!g_team_id || !g_me.id || g_weapon==WEAPON_OTHER)
   {
      scriptCall("doFire();");
      return;
   }

   // search for best target
   for(i=0; i<PLAYER_MAX; i++)
   {
      float       delta_yaw;
      float       delta_pitch;
      float       delta_mag;

      // ignore self as target
      if(!g_players[i].active || !g_players[i].team_id || *g_me.id==g_players[i].id)
         continue;
      if(!g_autoaim_teammates && g_players[i].team_id==g_team_id)
         continue;

      // get orientation to target
      aaGetOrientationToTarget(delta_yaw, delta_pitch, g_players[i].position_history[HISTORY_SIZE-1]);

      // find nearest target to center
      delta_mag = (float)sqrt(delta_yaw*delta_yaw + delta_pitch*delta_pitch/2.25f);
      if(ISNAN(delta_mag))
         delta_mag = 0.0f;
      if(delta_mag<mag) 
      {
         mag = delta_mag;
         final_yaw = delta_yaw;
         final_pitch = delta_pitch;
         final_index = i;
      }
   }

   // schedule autoaim if close to center
   if(mag<AA_SENSITIVITY)
   {
      sVertex3f   t;
      if(final_index >= 0)
      {
         aaGetNonLinearTarget(t, g_players[final_index]);
         aaGetOrientationToTarget(final_yaw, final_pitch, t);
      }
      // dont fire if using shocklance and out of range
      if(g_weapon==WEAPON_SHOCKLANCE)
      {
         float distance;
         distance = aaGetDistanceToTarget(t);
         if(distance < g_autoaim_sldistance)
            scriptCall("rotateFire(%0.8f, %0.8f);", final_yaw, final_pitch);
         else
            scriptMessage("target out of range: %0.1f [%0.1f]", distance, g_autoaim_sldistance);
      }
      else
         scriptCall("rotateFire(%0.8f, %0.8f);", final_yaw, final_pitch);
   }
   else
      scriptCall("doFire();");
}



// util
//-------------------------------------------
static void scriptCall( const char* command, ... )
{
   typedef void   (__cdecl *fScriptCall)( const char* command, int zero1, int zero2, int one );
   fScriptCall    script_call = (fScriptCall)CALL_SCRIPT;
   va_list        args;
   char           complete_command[BUF_STRING_SIZE];
   
   va_start(args, command);
   vsprintf(complete_command, command, args);
   va_end(args);

   script_call(complete_command, 0, 0, 1);
};

//-------------------------------------------
static void scriptMessage( const char* message, ... )
{
   va_list        args;
   char           complete_message[BUF_STRING_SIZE];
   float          delay;

   va_start(args, message);
   vsprintf(complete_message, message, args);
   va_end(args);

   // delay based on message length
   delay = strlen(complete_message)/7.0f;
   if(delay < 2.0f)
      delay = 2.0f;
   scriptCall("clientCmdBottomPrint( \"%s\", %f, 1 );", complete_message, delay);
};

//-------------------------------------------
static void scriptEcho( const char* message, ... )
{
   va_list        args;
   char           complete_message[BUF_STRING_SIZE];

   va_start(args, message);
   vsprintf(complete_message, message, args);
   va_end(args);

   scriptCall("echo(\"%s\");", complete_message);
};

//-------------------------------------------
static bool loadCrcs( void )
{
   FILE*          fp;
   char           line[BUF_LINE_SIZE];
   char           path[BUF_PATH_SIZE];
   int            i;

   // open
   sprintf(path,"%s\\%s", g_path, CRC_FILE);
   fp = fopen(path,"rt");
   if(fp == NULL)
      return false;

   // read
   for(i=0; fgets(line,BUF_LINE_SIZE,fp) && i<BUF_PATH_SIZE; )
   {
      if(line[i] == ';')
         continue;
      sscanf(line,"%x,%x %x,%x\n",
         &g_crc_map[i].crc_real[0], &g_crc_map[i].crc_real[1],
         &g_crc_map[i].crc_fake[0], &g_crc_map[i].crc_fake[1] );
      i++;
   }
   g_crc_count = i;

   // close
   fclose(fp);
   return true;
}


//-------------------------------------------
static bool checkFile( const char* file )
{
   FILE*          fp;
   char           path[BUF_PATH_SIZE];

   // open
   sprintf(path,"%s\\%s", g_path, file);
   fp = fopen(path,"r");
   if(fp == NULL)
      return false;
   // close
   fclose(fp);
   return true;
}

