// include
//--------------------------
#include <windows.h>
#include <stdarg.h>
#include <stdio.h>
#include <assert.h>

#include "share.h"
#include "gui.h"
#include "resource.h"


// globals
//--------------------------
extern bool start( HINSTANCE instance );
extern bool stop( void );

static HWND          g_hwnd = NULL;
static HINSTANCE     g_instance = NULL;
static int           g_state = GUI_IDLE;
static bool          g_is_app_started = false;


// local prototypes
//--------------------------
static bool guiStart( HINSTANCE instance );
static bool guiStop( void );
static DWORD WINAPI appThread( LPVOID lpParameter );
static LRESULT CALLBACK dialogProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);


// gui global functions
//--------------------------
bool guiPrint( const char* str, ... )
{
   va_list  args;
   int      index;
   char     message[BUF_STRING_SIZE];

   if(g_hwnd == NULL)
      return false;

   va_start(args, str);
   vsprintf(message, str, args);
   va_end(args);

   index = SendDlgItemMessage(g_hwnd, IDC_LIST, LB_ADDSTRING, (WPARAM)0, (LPARAM)(LPCTSTR)message);
   SendDlgItemMessage(g_hwnd, IDC_LIST, LB_SETCURSEL, (WPARAM)index, (LPARAM)0);

   return true;
}


//--------------------------
int guiState( void )
{
   return g_state;
}


//--------------------------
HWND guiGetHwnd( void )
{
   assert( g_hwnd != NULL );
   return g_hwnd;
}


// gui local functions
//--------------------------
static bool guiStart( HINSTANCE instance )
{
   g_instance = instance;
   g_state = GUI_RUNNING;
   DialogBox(g_instance, (LPCTSTR)IDD_LOADER, NULL, (DLGPROC)dialogProc);
   g_state = GUI_LOCAL_EXIT;
   g_hwnd = NULL;
   return true;
}

//--------------------------
static bool guiStop( void )
{
   while(g_is_app_started == true)
      Sleep(DELAY_SHORT);
   stop();
	EndDialog(g_hwnd, 0);
   g_hwnd = NULL;
   return true;
}


//--------------------------
static DWORD WINAPI appThread( LPVOID lpParameter )
{
   bool  r;
   g_is_app_started = true;
   r = start(g_instance);
   g_is_app_started = false;
   if(!r)
      SendMessage(g_hwnd, WM_COMMAND, WM_QUIT, 0);
   return 0;
}


//--------------------------
static LRESULT CALLBACK dialogProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
	case WM_INITDIALOG:
      {
         DWORD thread_id;
         g_hwnd = hDlg;
         SetClassLong(g_hwnd, GCL_HICON, (LONG)LoadIcon(g_instance, MAKEINTRESOURCE(IDI_ICON)) );
         if(CreateThread(NULL, 0, appThread, NULL, 0, &thread_id) == NULL)
         {
            g_state = GUI_LOCAL_EXIT;
            guiStop();
         }
      }
      return TRUE;

	case WM_COMMAND:
      if(g_state == GUI_RUNNING)
      {
		   switch(LOWORD(wParam))
         {
         case WM_QUIT:
         case WM_DESTROY:
         case WM_CLOSE:
            g_state = GUI_LOCAL_EXIT;
            guiStop();
            return TRUE;
		   }
      }
      return TRUE;
   }

   return FALSE;
}


// main
//--------------------------
int APIENTRY WinMain(HINSTANCE hinstance,
                     HINSTANCE prev_hinstance,
                     LPSTR     cmd_line,
                     int       cmd_show)
{
   if(!guiStart(hinstance))
      return 1;
   return 0;
}
