/*
   BUGS: dont hook at branch targets!!!
*/
   
// include
//--------------------------
#include <assert.h>
#include "share.h"
#include "List.h"
#include "Table.h"
#include "hook.h"
#include "x86.h"


// constants
//--------------------------
#define HOOK_PATCH_INCREMENT     100
#define HOOK_BACKUP_LENGTH       (OP_MAXSIZE*2)


// types
//--------------------------
struct sHook
{
   DWORD             hook_address;
   unsigned char*    patch_code;
   unsigned int      patch_code_length;
   unsigned int      patch_code_maxlength;
   unsigned char     backup_code[HOOK_BACKUP_LENGTH];
   unsigned int      backup_code_length;
   void*             handler;
};


// globals
//--------------------------
static Table<DWORD, sHook*>   g_hook_table;


// local functions
//--------------------------
bool doHook( sHook& hook );


// Hook functions
//--------------------------
bool hookSet( DWORD hook_address, void* hook_handler )
{
   DWORD    old_permissions;
   sHook*   hook;

   // check if exists
   hook = g_hook_table.get(hook_address);
   if(hook != NULL)
      return false;

   // add new hook
   hook = new sHook;
   hook->hook_address = hook_address;
   hook->patch_code = new unsigned char[HOOK_PATCH_INCREMENT];
   hook->patch_code_length = 0;
   hook->patch_code_maxlength = HOOK_PATCH_INCREMENT;
   if(!x86GetInstrBufferLength((unsigned char *)hook->hook_address, OP_BRANCH_SIZE, hook->backup_code_length))
      return false;

   memcpy(hook->backup_code, (char *)hook->hook_address, hook->backup_code_length);
   if( !VirtualProtect((void *)hook->hook_address, hook->backup_code_length, PAGE_EXECUTE_READWRITE, &old_permissions) )
      return false;
   hook->handler = hook_handler;
   g_hook_table.add(hook_address, hook);

   // make hook code changes
   if(!doHook(*hook))
      return false;

   return true;
}

//--------------------------
bool hookRemove( DWORD hook_address, void* hook_handler )
{
   sHook*   hook;

   hook = g_hook_table.get(hook_address);
   if(hook == NULL)
      return false;
   g_hook_table.remove(hook_address);
   memcpy((char *)hook->hook_address, hook->backup_code, hook->backup_code_length);
   delete[] hook->patch_code;
   delete   hook;

   return true;
}


// local functions
//--------------------------
static bool doHook( sHook& hook )
{
   int               offset = 0;
   unsigned char*    p_patch_code;

   // temporarily restore original code for thread safety
   memcpy((unsigned char *)hook.hook_address, hook.backup_code, hook.backup_code_length);

   // increase patch code buffer size if we're in danger of needing more
   if(hook.patch_code_length >= (hook.patch_code_maxlength-HOOK_BACKUP_LENGTH))
   {
      delete[] hook.patch_code;
      hook.patch_code_maxlength += HOOK_PATCH_INCREMENT;
      hook.patch_code = new unsigned char[hook.patch_code_maxlength];
   }
   p_patch_code = hook.patch_code;

   // PUSHAD
   // push general purpose registers
   *(unsigned char *)&p_patch_code[0] = OP_PUSHAD;
   p_patch_code++;

   // CALL [HOOK]
   // add call-to-hook_handler to patch code buffer
   offset = (DWORD)hook.handler - (((DWORD)&p_patch_code[0])+OP_BRANCH_SIZE);
   *(unsigned char *)&p_patch_code[0] = (unsigned char)OP_CALL;
   *(DWORD *)&p_patch_code[1] = offset;
   p_patch_code += OP_BRANCH_SIZE;
   
   // POPAD
   // pop general purpose registers
   *(unsigned char *)&p_patch_code[0] = OP_POPAD;
   p_patch_code++;

   // relocate hook-backup-code to patch code buffer
   if(!x86Relocate(hook.backup_code, hook.backup_code_length, hook.hook_address, p_patch_code, offset))
      return false;
   p_patch_code += offset;

   // JMP [HOOK+5]
   // add jump-back-to-after-hook code to patch code buffer
   offset = (hook.hook_address+hook.backup_code_length) - (((DWORD)&p_patch_code[0])+OP_BRANCH_SIZE);
   *(unsigned char *)&p_patch_code[0] = (unsigned char)OP_JMP;
   *(DWORD *)&p_patch_code[1] = offset;
   p_patch_code += OP_BRANCH_SIZE;

   // update patch code size
   hook.patch_code_length = (p_patch_code - hook.patch_code);

   // JMP [PATCH_CODE]
   // update jump-to-patch hook buffer code
   DWORD    hook_end = hook.hook_address+OP_BRANCH_SIZE;
   offset = (DWORD)hook.patch_code - hook_end;
   *(unsigned char *)hook.hook_address = (unsigned char)OP_JMP;
   *(int *)(hook.hook_address+1) = offset;

   // NOP
   // pad remaining code with no ops
   memset((void *)hook_end, OP_NOP, (unsigned int)(hook.backup_code_length-OP_BRANCH_SIZE));

   return true;
}
