// include
//--------------------------------------------
#include "m.h"
#include <math.h>


// vectors
//----------------------------------------------------
float vecMagnitude2( const sVertex2f& in_vector )
{
   return (float)(sqrt(in_vector.x*in_vector.x + in_vector.y*in_vector.y));
}

//----------------------------------------------------
float vecMagnitude3( const sVertex3f& a )
{
   return (float)sqrt(
      a.x*a.x +
      a.y*a.y +
      a.z*a.z);
}

//----------------------------------------------------
float vecAngle2( const sVertex2f& in_v1, const sVertex2f& in_v2 )
{
   float    dot_product;
   float    mag_product;
   float    angle;
   int      polarity;

   dot_product = in_v1.x*in_v2.x + in_v1.y*in_v2.y;
   mag_product = vecMagnitude2(in_v1) * vecMagnitude2(in_v2);
   polarity = (in_v1.x*in_v2.y - in_v1.y*in_v2.x)>=0?1:-1; // z of cross product
   angle = (float)acos(dot_product/mag_product) * polarity;
   if(ISNAN(angle))
      angle = 0.0f;
   return angle;
}

//----------------------------------------------------
void vecSubtract2( sVertex2f& out_v, const sVertex2f& in_v1, const sVertex2f& in_v2 )
{
   out_v.x = in_v1.x - in_v2.x;
   out_v.y = in_v1.y - in_v2.y;
}

//----------------------------------------------------
void vecSubtract3( sVertex3f& out_v, const sVertex3f& in_v1, const sVertex3f& in_v2 )
{
   out_v.x = in_v1.x - in_v2.x;
   out_v.y = in_v1.y - in_v2.y;
   out_v.z = in_v1.z - in_v2.z;
}


//----------------------------------------------------
float vecDirCos3( sVertex3f& out_a, const sVertex3f& v )
{
   float    m;

   m = vecMagnitude3( v );
   if(m == 0.0f)
      out_a.x = out_a.y = out_a.z = 0.0f;
   else
   {
      out_a.x = (float)acos(v.x/m);
      out_a.y = (float)acos(v.y/m);
      out_a.z = (float)acos(v.z/m);
   }

   return m;
}
